***********************************************************************************
                
	*** CLIMATE IMMOBILITY TRAPS: A HOUSEHOLD-LEVEL TEST ***    
                                                         
***********************************************************************************
                                                      
********************************************************************************************************************************************
********************************************************************************************************************************************
													  
**  POST-ESTIMATION CATE ANALYSIS  **   
**                                                            
**  AUTHORS: Marco Letta (Sapienza University of Rome), Pierluigi Montalbano (Sapienza University of Rome), Adriana Paolantonio (World Bank)
**                                                            
**  DATE: October 2023

********************************************************************************************************************************************
********************************************************************************************************************************************

*======================================================================================================*

* This do-file: takes the estimated treatment effects exported from R and analyzes conditional average treatment effects

* Final output: Figures 3 and 4 of the paper; Table A1
*======================================================================================================*
* CATE Analysis - Main
clear all
global dir "C:/Users/wb614536/OneDrive - WBG/Documents/Reproducibility/20231030-climate-immobility/files submitted/Reproducibility package/"
global output "${dir}/Outputs"
set varabbrev on
sysdir set PLUS "${dir}/ado"

import delimited "${dir}/Datasets/results_main.csv", // please change working directory to load this file
destring *, ignore("NA") replace
global vars "l1adapt l1share_tertiary l1hh_asset_index l1totcons l1spei_gs spei_gs"
g loci = tauhat - 1.965*sqrt
g hici = tauhat + 1.965*sqrt
g t = tauhat/sqrt
su t, d
g sig5 = abs(t)>=1.96
g sig10 = abs(t)>=1.65
label var tauhat "Estimated treatment effect (%)"
label var l1adapt "In situ adaptive capacity"
label var l1hh_asset_index "Asset index"
label var l1totcons "Total per adult equivalent consumption"
label var l1share_tertiary "Share of members with tertiary education"
label var l1spei_gs "Lagged GS-SPEI (reversed)"
label var spei_gs "GS-SPEI (reversed)"
global nbin = 10
*FIGURE 3*
binscatter tauhat l1adapt, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("In situ adaptive capacity", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1adapt, replace) graphregion(fcolor(white) color(white))
binscatter tauhat l1hh_asset_index, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("Asset index", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1hh_asset_index, replace) graphregion(fcolor(white) color(white))
binscatter tauhat l1totcons, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("Total consumption (per capita)", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1totcons, replace) graphregion(fcolor(white) color(white))
binscatter tauhat l1spei_gs, line(none) title("", size(medlarge))  m(t) n($nbin) xtitle("Lagged GS-SPEI (reversed)", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1spei_gs, replace) graphregion(fcolor(white) color(white))
binscatter tauhat l1share_tertiary, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("Sh. of members with tertiary education", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1share_tertiary, replace) graphregion(fcolor(white) color(white))
binscatter tauhat spei_gs, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("GS-SPEI (reversed)", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(spei_gs, replace) graphregion(fcolor(white) color(white))
graph combine l1hh_asset_index l1spei_gs l1adapt l1totcons l1share_tertiary spei_gs, ///
name(out, replace) rows(3) xsize(11) ysize(10) plotregion(margin(-1 0 -1 -1 )) imargin(vsmall) graphregion(fcolor(white) color(white)) // xsize(7) ysize(10)
graph export "${output}/Figure 3.emf", replace

*FIGURE 4*
* heatmaps
global nq = 4
cap drop q_*
gquantiles q_l1adapt =l1adapt, xtile nq($nq)
gquantiles q_l1hh_asset_index= l1hh_asset_index, xtile nq($nq)
gquantiles q_l1spei_gs = l1spei_gs, xtile nq($nq)
gquantiles q_spei_gs = spei_gs, xtile nq($nq)
label var tauhat "Predicted CATE"
heatplot tauhat q_l1adapt q_l1hh_asset_index, discrete(1) statistic(mean)  ylabel(1(1)4, labsize(vsmall)) xlabel(1(1)4, labsize(vsmall)) title("{bf:Panel A}", size(small)) /// 
 ytitle("In situ adaptive capacity", size(small))  xtitle("Asset index", size(small)) colors(magma, reverse gs) graphregion(fcolor(white) color(white)) leg(size(vsmall) subtitle("Treatment effect (%)", size(vsmall))) ///
name(adaptass, replace)

heatplot tauhat q_l1spei_gs q_l1adapt, discrete(1) statistic(mean)  ylabel(1(1)4, labsize(vsmall)) xlabel(1(1)4, labsize(vsmall))  title("{bf:Panel B}", size(small))  /// 
ytitle("Lagged GS-SPEI (reversed)", size(small))  xtitle("In situ adaptive capacity", size(small)) colors(magma, reverse gs) graphregion(fcolor(white) color(white)) leg(size(vsmall) subtitle("Treatment effect (%)", size(vsmall))) ///
name(adspei, replace)

heatplot tauhat q_l1spei_gs q_l1hh_asset_index, discrete(1) statistic(mean)  ylabel(1(1)4, labsize(vsmall)) xlabel(1(1)4, labsize(vsmall))  title("{bf:Panel C}", size(small))  /// 
ytitle("Lagged GS-SPEI (reversed)", size(small))  xtitle("Asset index", size(small)) colors(magma, reverse gs) graphregion(fcolor(white) color(white)) leg(size(vsmall) subtitle("Treatment effect (%)", size(vsmall))) ///
name(assspei, replace)

heatplot tauhat q_l1spei_gs q_spei_gs, discrete(1) statistic(mean)  ylabel(1(1)4, labsize(vsmall)) xlabel(1(1)4, labsize(vsmall))   title("{bf:Panel D}", size(small)) /// 
ytitle("Lagged GS-SPEI (reversed)", size(small))  xtitle("GS-SPEI (reversed)", size(small)) colors(magma, reverse gs) graphregion(fcolor(white) color(white)) leg(size(vsmall) subtitle("Treatment effect (%)", size(vsmall))) ///
name(spspei, replace)

graph combine adaptass adspei assspei spspei, ///
title("") name(out, replace) rows(2) plotregion(margin(-1 -1 -1 -1 )) imargin(med) graphregion(fcolor(white) color(white)) xsize(12) ysize(6)
graph export "${output}/Figure 4.emf", replace 

* Descriptive statistics
*TABLE A1*
estpost summ migrated spei_gs adapt share_tertiary hh_asset_index totcons age_head female_head hhsize nchildren nworkage houseshock manmadeshock
esttab ., cells("mean (fmt(a3)) sd (fmt(a3)) min (fmt(a3)) max (fmt(a3)) ") noobs compress, using "${output}/TableA1.rtf", replace

