***********************************************************************************
                
	*** CLIMATE IMMOBILITY TRAPS: A HOUSEHOLD-LEVEL TEST ***    
                                                         
***********************************************************************************
                                                      
********************************************************************************************************************************************
********************************************************************************************************************************************
													  
**  POST-ESTIMATION PLACEBO CATE ANALYSIS  **   
**                                                            
**  AUTHORS: Marco Letta (Sapienza University of Rome), Pierluigi Montalbano (Sapienza University of Rome), Adriana Paolantonio (World Bank)
**                                                            
**  DATE: October 2023

********************************************************************************************************************************************
********************************************************************************************************************************************

*===============================================================================================================*

* This do-file: takes the estimated placebo treatment effects exported from R and analyzes conditional average treatment effects for the placebo test

* Final output: Figure A2 of the paper

*===============================================================================================================*
* Placebo CATE Analysis
* change this path
global dir "C:/Users/wb614536/OneDrive - WBG/Documents/Reproducibility/20231030-climate-immobility/files submitted/Reproducibility package/"
global output "${dir}/Outputs"
sysdir set PLUS "${dir}/ado"

import delimited "${dir}/Datasets/results_placebo.csv", clear  // please change working directory to load this file
destring *, ignore("NA") replace
global vars "l1adapt l1share_tertiary l1hh_asset_index l1totcons l1spei_gs spei_gs"
g loci = tauhat - 1.965*sqrt
g hici = tauhat + 1.965*sqrt
g t = tauhat/sqrt
su t, d
g sig5 = abs(t)>=1.96
g sig10 = abs(t)>=1.65
label var tauhat "Estimated treatment effect (%)"
label var l1adapt "In situ adaptive capacity"
label var l1hh_asset_index "Asset index"
label var l1totcons "Total per adult equivalent consumption"
label var l1share_tertiary "Share of members with tertiary education"
label var l1spei_gs "Lagged placebo GS-SPEI (reversed)"
label var spei_gs "Placebo GS-SPEI (reversed)"
*replace tauhat = tauhat*100
global nbin = 10
*FIGURE A2*
binscatter tauhat l1adapt, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("In situ adaptive capacity", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1adapt, replace) graphregion(fcolor(white) color(white))
binscatter tauhat l1hh_asset_index, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("Asset index", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1hh_asset_index, replace) graphregion(fcolor(white) color(white))
binscatter tauhat l1totcons, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("Total consumption (per capita)", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1totcons, replace) graphregion(fcolor(white) color(white))
binscatter tauhat l1spei_gs, line(none) title("", size(medlarge))  m(t) n($nbin) xtitle("Lagged placebo GS-SPEI (reversed)", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1spei_gs, replace) graphregion(fcolor(white) color(white))
binscatter tauhat l1share_tertiary, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("Sh. of members with tertiary education", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(l1share_tertiary, replace) graphregion(fcolor(white) color(white))
binscatter tauhat spei_gs, line(none) title("", size(medlarge)) m(t) n($nbin) xtitle("Placebo GS-SPEI (reversed)", size(medsmall)) ytitle("Treatment effect (%)")  ylabels(-15 -10 -5 0 5 10 15) name(spei_gs, replace) graphregion(fcolor(white) color(white))
graph combine l1hh_asset_index l1spei_gs l1adapt l1totcons l1share_tertiary spei_gs, ///
name(out, replace) rows(3) xsize(11) ysize(10) plotregion(margin(-1 0 -1 -1 )) imargin(vsmall) graphregion(fcolor(white) color(white)) // xsize(7) ysize(10)
graph export "${output}/Figure A2.emf", replace 
